Global handle.l
Global DRIVER_NAME.s = "Kc2000Labs"
Global DRIVER_FILE_NAME.s ;= "DataLoader.X64";
Global DRIVER_NAME_FULL.s = "\\.\Kc2000Labs"

Structure PCIDEBUG_CONFREAD_INPUT
pci_address.l
pci_offset.l
EndStructure

Structure PCIDEBUG_MEMREAD_INPUT
address.q;l
unitsize.l
count.l
EndStructure

Procedure.l ctl_code(DeviceType, Function, Method, Access)
    ProcedureReturn ((DeviceType) << 16) | ((Access) << 14) | ((Function) << 2) | (Method)
EndProcedure

Procedure.l pciConfigRead(pci_address.l, reg_address.l, *datas, size.l)
	buf.PCIDEBUG_CONFREAD_INPUT
	ReturnedLength.l
	IoctlResult.l
	buf\pci_address = pci_address
	buf\pci_offset = reg_address
	IOCTL_PCI_READ_CONF.l = ctl_code(40000,$940,0,1)

	IoctlResult = DeviceIoControl_(handle, IOCTL_PCI_READ_CONF, @buf, SizeOf(buf), *datas, size, @ReturnedLength, #Null)
	If ( Not IoctlResult)                            ;/* Did the IOCTL succeed? */
		ProcedureReturn GetLastError_()
	EndIf

	ProcedureReturn 0;
EndProcedure

Procedure.l PciConfigReadLong(pci_address.l, reg_address.l)
	ret.l
	If pciConfigRead(pci_address, reg_address, @ret, SizeOf(ret)) 
		;/* error */
		ProcedureReturn $ffffffff
	EndIf
	ProcedureReturn ret
EndProcedure

Procedure.l ReadMSR(ulECX.l, *ulEAX, *ulEDX)
	ReturnedLength.l
 	IoctlResult.l
	buf.s{8}
	IOCTL_READ_MSR.l = ctl_code(40000,$981,0,0)

	IoctlResult = DeviceIoControl_(handle, IOCTL_READ_MSR, @ulECX, SizeOf(ulECX), @buf, SizeOf(buf), @ReturnedLength, #Null)
	CopyMemory(@buf,*ulEAX,4)
	CopyMemory(@buf+4,*ulEDX,4)
	
	ProcedureReturn IoctlResult
EndProcedure

Procedure.c PciConfigReadChar(pci_address.l, reg_address.l)
	ret.c
	If pciConfigRead(pci_address, reg_address, @ret, SizeOf(ret))
		ProcedureReturn $ff
	EndIf
	ProcedureReturn ret
EndProcedure

Procedure.l MemReadBlock(address.l, *datas, count.l, unitsize.l)
	param.PCIDEBUG_MEMREAD_INPUT
	IoctlResult.l
	ReturnedLength.l
	size.l
	param\address = address
	param\unitsize = unitsize
	param\count = count
	size = param\unitsize * param\count
	IOCTL_PCI_READ_MEM.l = ctl_code(40000,$920,0,1)
    IoctlResult = DeviceIoControl_(handle, IOCTL_PCI_READ_MEM, @param, SizeOf(param), *datas, size, @ReturnedLength, #Null)
    If IoctlResult And ReturnedLength = size
		ProcedureReturn param\count
	EndIf
	ProcedureReturn 0
EndProcedure

Procedure.l MemReadBlockChar(address.l, *datas, count.l)
	ProcedureReturn MemReadBlock(address, *datas, count, 1)
EndProcedure

Procedure.l OpenDriver()
	handle = CreateFile_(@DRIVER_NAME_FULL, #GENERIC_READ | #GENERIC_WRITE, #FILE_SHARE_READ | #FILE_SHARE_WRITE, 0, #OPEN_EXISTING, #FILE_ATTRIBUTE_NORMAL, 0)
	If handle = #INVALID_HANDLE_VALUE
        ProcedureReturn 0
    EndIf
	ProcedureReturn 1
EndProcedure

Procedure.l LoadDriver(filename.s, drivername.s)
    hSCManager.l
    hService.l
    serviceStatus.SERVICE_STATUS
	ret.l


	hSCManager = OpenSCManager_(0, 0, #SC_MANAGER_ALL_ACCESS)
	If Not hSCManager
		ProcedureReturn 0
	EndIf

	hService = OpenService_(hSCManager, @drivername, #SERVICE_ALL_ACCESS)
	If hService
		ControlService_(hService, #SERVICE_CONTROL_STOP, @ServiceStatus)
		DeleteService_(hService)
		CloseServiceHandle_(hService)
	EndIf

	hService = CreateService_(hSCManager, @drivername, @drivername, #SERVICE_ALL_ACCESS, #SERVICE_KERNEL_DRIVER, #SERVICE_DEMAND_START, #SERVICE_ERROR_NORMAL, @filename, #Null, #Null, #Null, #Null, #Null)
	ret = 0
    If hService
		ret = StartService_(hService, 0, 0)
		CloseServiceHandle_(hService)
	EndIf
	CloseServiceHandle_(hSCManager)
	ProcedureReturn ret
EndProcedure


Procedure.l UnloadDriver(drivername.s)
	;SC_HANDLE	
	hSCManager.l
	hService.l
	serviceStatus.SERVICE_STATUS
	ret.l = 0
	
    hSCManager = OpenSCManager_(0, 0, #SC_MANAGER_ALL_ACCESS)
	If hSCManager
		ProcedureReturn 0
	EndIf
    
    hService = OpenService_(hSCManager, @drivername, #SERVICE_ALL_ACCESS)
	If hService
	    ret = ControlService_(hService, #SERVICE_CONTROL_STOP, @serviceStatus)
	    If ret = 1
	        ret = DeleteService_(hService)
		EndIf
	    CloseServiceHandle_(hService)
	EndIf

    CloseServiceHandle_(hSCManager)

    ProcedureReturn ret
EndProcedure

Procedure.l GetRefCount(hHandle.l)
	dwRefCount.l = -1
	ReturnedLength.l: IoctlResult.l
	IOCTL_GET_REFCONT.l = ctl_code(40000,$999,0,0)

	If hHandle
		IoctlResult = DeviceIoControl_(hHandle, IOCTL_GET_REFCONT, #Null, 0, @dwRefCount, SizeOf(dwRefCount), @ReturnedLength, #Null)
		If Not IoctlResult
			dwRefCount = -1
		EndIf
	EndIf

	ProcedureReturn dwRefCount
EndProcedure

Procedure close()
	dwRefCount.l = GetRefCount(handle)

	If handle
		CloseHandle_(handle)
	EndIf
	If dwRefCount = 1
		UnloadDriver(DRIVER_NAME)
	EndIf
	handle = 0
EndProcedure


Procedure.l initialize(DRIVER_FILE_NAME.s)
	ret.l
	path.s
	If OpenDriver()
		ProcedureReturn 0; no error
	EndIf
	path = GetPathPart(ProgramFilename()) + DRIVER_FILE_NAME
	ret = LoadDriver(path, DRIVER_NAME)
	If ret = 0
		ProcedureReturn 2; error
	EndIf

	If OpenDriver()
		ProcedureReturn 0;no error
	EndIf

	ProcedureReturn 3;error
EndProcedure

Procedure.l LoadSYS()
	ii.l: res.l
	For ii=0 To 16
		res = initialize(DRIVER_FILE_NAME)
		If res = 0
			;Debug "handle = " + Str(handle)
			ProcedureReturn 1	;Success
		EndIf
		Sleep_(100)
	Next
	msg.s = "            " + DRIVER_FILE_NAME + " ." +Chr(10)+ "           ."+Chr(10)+Chr(10)+"An  error  occured  when  opening  the  driver  " + DRIVER_FILE_NAME + " ."+Chr(10)+"The  program  will  not  run  properly."
	MessageRequester(" - Error",msg)
	ProcedureReturn 0
EndProcedure

Procedure IsWin64(); is the OS 64;
  Protected Info.SYSTEM_INFO
  GetNativeSystemInfo(Info)
  If info\wProcessorArchitecture
    ProcedureReturn #True
  EndIf
EndProcedure 

Procedure LoadDeviceDriver()
Select OSVersion()
	Case #PB_OS_Windows_95, #PB_OS_Windows_98, #PB_OS_Windows_ME
		MessageRequester("Unsupported OS","Your OS is not supported.")
	Default
		If IsWin64() = 1
			DRIVER_FILE_NAME = "DataLoader.X64"
		Else
			DRIVER_FILE_NAME = "DataLoader.NT"
		EndIf
		DrvStatus = LoadSYS()
EndSelect
EndProcedure
; IDE Options = PureBasic 4.31 (Windows - x86)
; CursorPosition = 93
; FirstLine = 76
; Folding = ---
; EnableXP
; Executable = pci.exe